@@ -492,6 +492,163 @@ TIntermTyped* TIntermConstantUnion::fold(TOperator op, const TType& returnType)
 
     // Process component-wise operations
     for (int i = 0; i < objectSize; i++) {
+        // First read the value and convert to i64/u64/f64/bool, then convert
+        // to the destination type (still 64b), then convert down to the
+        // destination size.
+        if (IsOpNumericConv(op)) {
+            enum ConvType { CONV_FLOAT, CONV_INT, CONV_UINT, CONV_BOOL };
+            ConvType srcType = CONV_UINT, dstType = CONV_UINT;
+            double valf = 0.0;
+            uint64_t valu = 0;
+            int64_t vali = 0;
+            bool valb = false;
+            switch (getType().getBasicType()) {
+            case EbtDouble:
+            case EbtFloat16:
+            case EbtFloat:
+                valf = unionArray[i].getDConst();
+                srcType = CONV_FLOAT;
+                break;
+            case EbtInt8:
+                vali = unionArray[i].getI8Const();
+                srcType = CONV_INT;
+                break;
+            case EbtInt16:
+                vali = unionArray[i].getI16Const();
+                srcType = CONV_INT;
+                break;
+            case EbtInt:
+                vali = unionArray[i].getIConst();
+                srcType = CONV_INT;
+                break;
+            case EbtInt64:
+                vali = unionArray[i].getI64Const();
+                srcType = CONV_INT;
+                break;
+            case EbtUint8:
+                valu = unionArray[i].getU8Const();
+                srcType = CONV_UINT;
+                break;
+            case EbtUint16:
+                valu = unionArray[i].getU16Const();
+                srcType = CONV_UINT;
+                break;
+            case EbtUint:
+                valu = unionArray[i].getUConst();
+                srcType = CONV_UINT;
+                break;
+            case EbtUint64:
+                valu = unionArray[i].getU64Const();
+                srcType = CONV_UINT;
+                break;
+            case EbtBool:
+                valb = unionArray[i].getBConst();
+                srcType = CONV_BOOL;
+                break;
+            default:
+                assert(0);
+                break;
+            }
+
+            switch (returnType.getBasicType()) {
+            case EbtDouble:
+            case EbtFloat16:
+            case EbtFloat:
+                dstType = CONV_FLOAT;
+                break;
+            case EbtInt8:
+            case EbtInt16:
+            case EbtInt:
+            case EbtInt64:
+                dstType = CONV_INT;
+                break;
+            case EbtUint8:
+            case EbtUint16:
+            case EbtUint:
+            case EbtUint64:
+                dstType = CONV_UINT;
+                break;
+            case EbtBool:
+                dstType = CONV_BOOL;
+                break;
+            default:
+                assert(0);
+                break;
+            }
+            if (dstType == CONV_BOOL) {
+                switch (srcType) {
+                case CONV_FLOAT:
+                    valb = (valf != 0.0); break;
+                case CONV_INT:
+                    valb = (vali != 0.0); break;
+                case CONV_UINT:
+                    valb = (valu != 0.0); break;
+                default:
+                    break;
+                }
+            } else if (dstType == CONV_FLOAT) {
+                switch (srcType) {
+                case CONV_BOOL:
+                    valf = (double)valb; break;
+                case CONV_INT:
+                    valf = (double)vali; break;
+                case CONV_UINT:
+                    valf = (double)valu; break;
+                default:
+                    break;
+                }
+            } else if (dstType == CONV_INT) {
+                switch (srcType) {
+                case CONV_BOOL:
+                    vali = (int64_t)valb; break;
+                case CONV_FLOAT:
+                    vali = (int64_t)valf; break;
+                case CONV_UINT:
+                    vali = (int64_t)valu; break;
+                default:
+                    break;
+                }
+            } else if (dstType == CONV_UINT) {
+                switch (srcType) {
+                case CONV_BOOL:
+                    valu = (uint64_t)valb; break;
+                case CONV_FLOAT:
+                    valu = (uint64_t)valf; break;
+                case CONV_INT:
+                    valu = (uint64_t)vali; break;
+                default:
+                    break;
+                }
+            }
+            switch (returnType.getBasicType()) {
+            case EbtDouble:
+            case EbtFloat16:
+            case EbtFloat:
+                newConstArray[i].setDConst(valf); break;
+            case EbtInt8:
+                newConstArray[i].setI8Const(static_cast<int8_t>(vali)); break;
+            case EbtInt16:
+                newConstArray[i].setI16Const(static_cast<int16_t>(vali)); break;
+            case EbtInt:
+                newConstArray[i].setIConst(static_cast<int32_t>(vali)); break;
+            case EbtInt64:
+                newConstArray[i].setI64Const(vali); break;
+            case EbtUint8:
+                newConstArray[i].setU8Const(static_cast<uint8_t>(valu)); break;
+            case EbtUint16:
+                newConstArray[i].setU16Const(static_cast<uint16_t>(valu)); break;
+            case EbtUint:
+                newConstArray[i].setUConst(static_cast<uint32_t>(valu)); break;
+            case EbtUint64:
+                newConstArray[i].setU64Const(valu); break;
+            case EbtBool:
+                newConstArray[i].setBConst(valb); break;
+            default:
+                assert(0);
+                break;
+            }
+            continue;
+        }
         switch (op) {
         case EOpNegative:
             switch (getType().getBasicType()) {
@@ -509,7 +666,11 @@ TIntermTyped* TIntermConstantUnion::fold(TOperator op, const TType& returnType)
             case EbtUint8: newConstArray[i].setU8Const(static_cast<unsigned int>(-static_cast<signed int>(unionArray[i].getU8Const())));  break;
             case EbtInt16: newConstArray[i].setI16Const(-unionArray[i].getI16Const()); break;
             case EbtUint16:newConstArray[i].setU16Const(static_cast<unsigned int>(-static_cast<signed int>(unionArray[i].getU16Const())));  break;
-            case EbtInt64: newConstArray[i].setI64Const(-unionArray[i].getI64Const()); break;
+            case EbtInt64: {
+                int64_t i64val = unionArray[i].getI64Const();
+                newConstArray[i].setI64Const(i64val == INT64_MIN ? INT64_MIN : -i64val);
+                break;
+            }
             case EbtUint64: newConstArray[i].setU64Const(static_cast<unsigned long long>(-static_cast<long long>(unionArray[i].getU64Const())));  break;
             default:
                 return nullptr;
@@ -639,277 +800,6 @@ TIntermTyped* TIntermConstantUnion::fold(TOperator op, const TType& returnType)
             break;
         }
 
-        case EOpConvIntToBool:
-            newConstArray[i].setBConst(unionArray[i].getIConst() != 0); break;
-        case EOpConvUintToBool:
-            newConstArray[i].setBConst(unionArray[i].getUConst() != 0); break;
-        case EOpConvBoolToInt:
-            newConstArray[i].setIConst(unionArray[i].getBConst()); break;
-        case EOpConvBoolToUint:
-            newConstArray[i].setUConst(unionArray[i].getBConst()); break;
-        case EOpConvIntToUint:
-            newConstArray[i].setUConst(unionArray[i].getIConst()); break;
-        case EOpConvUintToInt:
-            newConstArray[i].setIConst(unionArray[i].getUConst()); break;
-
-        case EOpConvFloatToBool:
-        case EOpConvDoubleToBool:
-            newConstArray[i].setBConst(unionArray[i].getDConst() != 0); break;
-
-        case EOpConvBoolToFloat:
-        case EOpConvBoolToDouble:
-            newConstArray[i].setDConst(unionArray[i].getBConst()); break;
-
-        case EOpConvIntToFloat:
-        case EOpConvIntToDouble:
-            newConstArray[i].setDConst(unionArray[i].getIConst()); break;
-
-        case EOpConvUintToFloat:
-        case EOpConvUintToDouble:
-            newConstArray[i].setDConst(unionArray[i].getUConst()); break;
-
-        case EOpConvDoubleToFloat:
-        case EOpConvFloatToDouble:
-            newConstArray[i].setDConst(unionArray[i].getDConst()); break;
-
-        case EOpConvFloatToUint:
-        case EOpConvDoubleToUint:
-            newConstArray[i].setUConst(static_cast<unsigned int>(unionArray[i].getDConst())); break;
-
-        case EOpConvFloatToInt:
-        case EOpConvDoubleToInt:
-            newConstArray[i].setIConst(static_cast<int>(unionArray[i].getDConst())); break;
-
-        case EOpConvInt8ToBool:
-            newConstArray[i].setBConst(unionArray[i].getI8Const() != 0); break;
-        case EOpConvUint8ToBool:
-            newConstArray[i].setBConst(unionArray[i].getU8Const() != 0); break;
-        case EOpConvInt16ToBool:
-            newConstArray[i].setBConst(unionArray[i].getI16Const() != 0); break;
-        case EOpConvUint16ToBool:
-            newConstArray[i].setBConst(unionArray[i].getU16Const() != 0); break;
-        case EOpConvInt64ToBool:
-            newConstArray[i].setBConst(unionArray[i].getI64Const() != 0); break;
-        case EOpConvUint64ToBool:
-            newConstArray[i].setBConst(unionArray[i].getU64Const() != 0); break;
-        case EOpConvFloat16ToBool:
-            newConstArray[i].setBConst(unionArray[i].getDConst() != 0); break;
-
-        case EOpConvBoolToInt8:
-            newConstArray[i].setI8Const(unionArray[i].getBConst()); break;
-        case EOpConvBoolToUint8:
-            newConstArray[i].setU8Const(unionArray[i].getBConst()); break;
-        case EOpConvBoolToInt16:
-            newConstArray[i].setI16Const(unionArray[i].getBConst()); break;
-        case EOpConvBoolToUint16:
-            newConstArray[i].setU16Const(unionArray[i].getBConst()); break;
-        case EOpConvBoolToInt64:
-            newConstArray[i].setI64Const(unionArray[i].getBConst()); break;
-        case EOpConvBoolToUint64:
-            newConstArray[i].setU64Const(unionArray[i].getBConst()); break;
-        case EOpConvBoolToFloat16:
-            newConstArray[i].setDConst(unionArray[i].getBConst()); break;
-
-        case EOpConvInt8ToInt16:
-            newConstArray[i].setI16Const(unionArray[i].getI8Const()); break;
-        case EOpConvInt8ToInt:
-            newConstArray[i].setIConst(unionArray[i].getI8Const()); break;
-        case EOpConvInt8ToInt64:
-            newConstArray[i].setI64Const(unionArray[i].getI8Const()); break;
-        case EOpConvInt8ToUint8:
-            newConstArray[i].setU8Const(unionArray[i].getI8Const()); break;
-        case EOpConvInt8ToUint16:
-            newConstArray[i].setU16Const(unionArray[i].getI8Const()); break;
-        case EOpConvInt8ToUint:
-            newConstArray[i].setUConst(unionArray[i].getI8Const()); break;
-        case EOpConvInt8ToUint64:
-            newConstArray[i].setU64Const(unionArray[i].getI8Const()); break;
-        case EOpConvUint8ToInt8:
-            newConstArray[i].setI8Const(unionArray[i].getU8Const()); break;
-        case EOpConvUint8ToInt16:
-            newConstArray[i].setI16Const(unionArray[i].getU8Const()); break;
-        case EOpConvUint8ToInt:
-            newConstArray[i].setIConst(unionArray[i].getU8Const()); break;
-        case EOpConvUint8ToInt64:
-            newConstArray[i].setI64Const(unionArray[i].getU8Const()); break;
-        case EOpConvUint8ToUint16:
-            newConstArray[i].setU16Const(unionArray[i].getU8Const()); break;
-        case EOpConvUint8ToUint:
-            newConstArray[i].setUConst(unionArray[i].getU8Const()); break;
-        case EOpConvUint8ToUint64:
-            newConstArray[i].setU64Const(unionArray[i].getU8Const()); break;
-        case EOpConvInt8ToFloat16:
-            newConstArray[i].setDConst(unionArray[i].getI8Const()); break;
-        case EOpConvInt8ToFloat:
-            newConstArray[i].setDConst(unionArray[i].getI8Const()); break;
-        case EOpConvInt8ToDouble:
-            newConstArray[i].setDConst(unionArray[i].getI8Const()); break;
-        case EOpConvUint8ToFloat16:
-            newConstArray[i].setDConst(unionArray[i].getU8Const()); break;
-        case EOpConvUint8ToFloat:
-            newConstArray[i].setDConst(unionArray[i].getU8Const()); break;
-        case EOpConvUint8ToDouble:
-            newConstArray[i].setDConst(unionArray[i].getU8Const()); break;
-
-        case EOpConvInt16ToInt8:
-            newConstArray[i].setI8Const(static_cast<signed char>(unionArray[i].getI16Const())); break;
-        case EOpConvInt16ToInt:
-            newConstArray[i].setIConst(unionArray[i].getI16Const()); break;
-        case EOpConvInt16ToInt64:
-            newConstArray[i].setI64Const(unionArray[i].getI16Const()); break;
-        case EOpConvInt16ToUint8:
-            newConstArray[i].setU8Const(static_cast<unsigned char>(unionArray[i].getI16Const())); break;
-        case EOpConvInt16ToUint16:
-            newConstArray[i].setU16Const(unionArray[i].getI16Const()); break;
-        case EOpConvInt16ToUint:
-            newConstArray[i].setUConst(unionArray[i].getI16Const()); break;
-        case EOpConvInt16ToUint64:
-            newConstArray[i].setU64Const(unionArray[i].getI16Const()); break;
-        case EOpConvUint16ToInt8:
-            newConstArray[i].setI8Const(static_cast<signed char>(unionArray[i].getU16Const())); break;
-        case EOpConvUint16ToInt16:
-            newConstArray[i].setI16Const(unionArray[i].getU16Const()); break;
-        case EOpConvUint16ToInt:
-            newConstArray[i].setIConst(unionArray[i].getU16Const()); break;
-        case EOpConvUint16ToInt64:
-            newConstArray[i].setI64Const(unionArray[i].getU16Const()); break;
-        case EOpConvUint16ToUint8:
-            newConstArray[i].setU8Const(static_cast<unsigned char>(unionArray[i].getU16Const())); break;
-
-        case EOpConvUint16ToUint:
-            newConstArray[i].setUConst(unionArray[i].getU16Const()); break;
-        case EOpConvUint16ToUint64:
-            newConstArray[i].setU64Const(unionArray[i].getU16Const()); break;
-        case EOpConvInt16ToFloat16:
-            newConstArray[i].setDConst(unionArray[i].getI16Const()); break;
-        case EOpConvInt16ToFloat:
-            newConstArray[i].setDConst(unionArray[i].getI16Const()); break;
-        case EOpConvInt16ToDouble:
-            newConstArray[i].setDConst(unionArray[i].getI16Const()); break;
-        case EOpConvUint16ToFloat16:
-            newConstArray[i].setDConst(unionArray[i].getU16Const()); break;
-        case EOpConvUint16ToFloat:
-            newConstArray[i].setDConst(unionArray[i].getU16Const()); break;
-        case EOpConvUint16ToDouble:
-            newConstArray[i].setDConst(unionArray[i].getU16Const()); break;
-
-        case EOpConvIntToInt8:
-            newConstArray[i].setI8Const((signed char)unionArray[i].getIConst()); break;
-        case EOpConvIntToInt16:
-            newConstArray[i].setI16Const((signed short)unionArray[i].getIConst()); break;
-        case EOpConvIntToInt64:
-            newConstArray[i].setI64Const(unionArray[i].getIConst()); break;
-        case EOpConvIntToUint8:
-            newConstArray[i].setU8Const((unsigned char)unionArray[i].getIConst()); break;
-        case EOpConvIntToUint16:
-            newConstArray[i].setU16Const((unsigned char)unionArray[i].getIConst()); break;
-        case EOpConvIntToUint64:
-            newConstArray[i].setU64Const(unionArray[i].getIConst()); break;
-
-        case EOpConvUintToInt8:
-            newConstArray[i].setI8Const((signed char)unionArray[i].getUConst()); break;
-        case EOpConvUintToInt16:
-            newConstArray[i].setI16Const((signed short)unionArray[i].getUConst()); break;
-        case EOpConvUintToInt64:
-            newConstArray[i].setI64Const(unionArray[i].getUConst()); break;
-        case EOpConvUintToUint8:
-            newConstArray[i].setU8Const((unsigned char)unionArray[i].getUConst()); break;
-        case EOpConvUintToUint16:
-            newConstArray[i].setU16Const((unsigned short)unionArray[i].getUConst()); break;
-        case EOpConvUintToUint64:
-            newConstArray[i].setU64Const(unionArray[i].getUConst()); break;
-        case EOpConvIntToFloat16:
-            newConstArray[i].setDConst(unionArray[i].getIConst()); break;
-        case EOpConvUintToFloat16:
-            newConstArray[i].setDConst(unionArray[i].getUConst()); break;
-        case EOpConvInt64ToInt8:
-            newConstArray[i].setI8Const(static_cast<signed char>(unionArray[i].getI64Const())); break;
-        case EOpConvInt64ToInt16:
-            newConstArray[i].setI16Const(static_cast<signed short>(unionArray[i].getI64Const())); break;
-        case EOpConvInt64ToInt:
-            newConstArray[i].setIConst(static_cast<int>(unionArray[i].getI64Const())); break;
-        case EOpConvInt64ToUint8:
-            newConstArray[i].setU8Const(static_cast<unsigned char>(unionArray[i].getI64Const())); break;
-        case EOpConvInt64ToUint16:
-            newConstArray[i].setU16Const(static_cast<unsigned short>(unionArray[i].getI64Const())); break;
-        case EOpConvInt64ToUint:
-            newConstArray[i].setUConst(static_cast<unsigned int>(unionArray[i].getI64Const())); break;
-        case EOpConvInt64ToUint64:
-            newConstArray[i].setU64Const(unionArray[i].getI64Const()); break;
-        case EOpConvUint64ToInt8:
-            newConstArray[i].setI8Const(static_cast<signed char>(unionArray[i].getU64Const())); break;
-        case EOpConvUint64ToInt16:
-            newConstArray[i].setI16Const(static_cast<signed short>(unionArray[i].getU64Const())); break;
-        case EOpConvUint64ToInt:
-            newConstArray[i].setIConst(static_cast<int>(unionArray[i].getU64Const())); break;
-        case EOpConvUint64ToInt64:
-            newConstArray[i].setI64Const(unionArray[i].getU64Const()); break;
-        case EOpConvUint64ToUint8:
-            newConstArray[i].setU8Const(static_cast<unsigned char>(unionArray[i].getU64Const())); break;
-        case EOpConvUint64ToUint16:
-            newConstArray[i].setU16Const(static_cast<unsigned short>(unionArray[i].getU64Const())); break;
-        case EOpConvUint64ToUint:
-            newConstArray[i].setUConst(static_cast<unsigned int>(unionArray[i].getU64Const())); break;
-        case EOpConvInt64ToFloat16:
-            newConstArray[i].setDConst(static_cast<double>(unionArray[i].getI64Const())); break;
-        case EOpConvInt64ToFloat:
-            newConstArray[i].setDConst(static_cast<double>(unionArray[i].getI64Const())); break;
-        case EOpConvInt64ToDouble:
-            newConstArray[i].setDConst(static_cast<double>(unionArray[i].getI64Const())); break;
-        case EOpConvUint64ToFloat16:
-            newConstArray[i].setDConst(static_cast<double>(unionArray[i].getU64Const())); break;
-        case EOpConvUint64ToFloat:
-            newConstArray[i].setDConst(static_cast<double>(unionArray[i].getU64Const())); break;
-        case EOpConvUint64ToDouble:
-            newConstArray[i].setDConst(static_cast<double>(unionArray[i].getU64Const())); break;
-        case EOpConvFloat16ToInt8:
-            newConstArray[i].setI8Const(static_cast<signed char>(unionArray[i].getDConst())); break;
-        case EOpConvFloat16ToInt16:
-            newConstArray[i].setI16Const(static_cast<signed short>(unionArray[i].getDConst())); break;
-        case EOpConvFloat16ToInt:
-            newConstArray[i].setIConst(static_cast<int>(unionArray[i].getDConst())); break;
-        case EOpConvFloat16ToInt64:
-            newConstArray[i].setI64Const(static_cast<long long>(unionArray[i].getDConst())); break;
-        case EOpConvFloat16ToUint8:
-            newConstArray[i].setU8Const(static_cast<unsigned char>(unionArray[i].getDConst())); break;
-        case EOpConvFloat16ToUint16:
-            newConstArray[i].setU16Const(static_cast<unsigned short>(unionArray[i].getDConst())); break;
-        case EOpConvFloat16ToUint:
-            newConstArray[i].setUConst(static_cast<unsigned int>(unionArray[i].getDConst())); break;
-        case EOpConvFloat16ToUint64:
-            newConstArray[i].setU64Const(static_cast<unsigned long long>(unionArray[i].getDConst())); break;
-        case EOpConvFloat16ToFloat:
-            newConstArray[i].setDConst(unionArray[i].getDConst()); break;
-        case EOpConvFloat16ToDouble:
-            newConstArray[i].setDConst(unionArray[i].getDConst()); break;
-        case EOpConvFloatToInt8:
-            newConstArray[i].setI8Const(static_cast<signed char>(unionArray[i].getDConst())); break;
-        case EOpConvFloatToInt16:
-            newConstArray[i].setI16Const(static_cast<signed short>(unionArray[i].getDConst())); break;
-        case EOpConvFloatToInt64:
-            newConstArray[i].setI64Const(static_cast<long long>(unionArray[i].getDConst())); break;
-        case EOpConvFloatToUint8:
-            newConstArray[i].setU8Const(static_cast<unsigned char>(unionArray[i].getDConst())); break;
-        case EOpConvFloatToUint16:
-            newConstArray[i].setU16Const(static_cast<unsigned short>(unionArray[i].getDConst())); break;
-        case EOpConvFloatToUint64:
-            newConstArray[i].setU64Const(static_cast<unsigned long long>(unionArray[i].getDConst())); break;
-        case EOpConvFloatToFloat16:
-            newConstArray[i].setDConst(unionArray[i].getDConst()); break;
-        case EOpConvDoubleToInt8:
-            newConstArray[i].setI8Const(static_cast<signed char>(unionArray[i].getDConst())); break;
-        case EOpConvDoubleToInt16:
-            newConstArray[i].setI16Const(static_cast<signed short>(unionArray[i].getDConst())); break;
-        case EOpConvDoubleToInt64:
-            newConstArray[i].setI64Const(static_cast<long long>(unionArray[i].getDConst())); break;
-        case EOpConvDoubleToUint8:
-            newConstArray[i].setU8Const(static_cast<unsigned char>(unionArray[i].getDConst())); break;
-        case EOpConvDoubleToUint16:
-            newConstArray[i].setU16Const(static_cast<unsigned short>(unionArray[i].getDConst())); break;
-        case EOpConvDoubleToUint64:
-            newConstArray[i].setU64Const(static_cast<unsigned long long>(unionArray[i].getDConst())); break;
-        case EOpConvDoubleToFloat16:
-            newConstArray[i].setDConst(unionArray[i].getDConst()); break;
         case EOpConvPtrToUint64:
         case EOpConvUint64ToPtr:
         case EOpConstructReference:
@@ -1011,6 +901,12 @@ TIntermTyped* TIntermediate::fold(TIntermAggregate* aggrNode)
         objectSize = std::max(children[0]->getAsTyped()->getType().getVectorSize(),
                               children[2]->getAsTyped()->getType().getVectorSize());
         break;
+    case EOpMul:
+        {
+        TIntermConstantUnion* left = children[0]->getAsConstantUnion();
+        TIntermConstantUnion* right = children[1]->getAsConstantUnion();
+        return left->fold(EOpMul, right);
+        }
     default:
         return aggrNode;
     }
