@@ -1,7 +1,7 @@
-// Copyright (C) 2021 The Qt Company Ltd.
+// Copyright (C) 2023 The Qt Company Ltd.
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
-#include "qshaderbaker_p.h"
+#include "qshaderbaker.h"
 #include "qspirvcompiler_p.h"
 #include "qspirvshader_p.h"
 #include <QFileInfo>
@@ -12,12 +12,23 @@ QT_BEGIN_NAMESPACE
 
 /*!
     \class QShaderBaker
-    \internal
     \inmodule QtShaderTools
+    \since 6.6
 
     \brief Compiles a GLSL/Vulkan shader into SPIR-V, translates into other
     shading languages, and gathers reflection metadata.
 
+    \warning QShaderBaker, just like the QRhi family of classes in the Qt Gui
+    module, including QShader and QShaderDescription, offers limited
+    compatibility guarantees. There are no source or binary compatibility
+    guarantees for these classes, meaning the API is only guaranteed to work
+    with the Qt version the application was developed against. Source
+    incompatible changes are however aimed to be kept at a minimum and will only
+    be made in minor releases (6.7, 6.8, and so on). To use this class in an
+    application, link to \c{Qt::ShaderToolsPrivate} (if using CMake), and
+    include the headers with the \c rhi prefix, for example
+    \c{#include <rhi/qshaderbaker.h>}.
+
     QShaderBaker takes a graphics (vertex, fragment, etc.) or compute shader,
     and produces multiple - either source or bytecode - variants of it,
     together with reflection information. The results are represented by a
@@ -25,12 +36,12 @@ QT_BEGIN_NAMESPACE
     and deserialization.
 
     \note Applications and libraries are recommended to avoid using this class
-    directly. Rather, all Qt users are encouraged to rely on offline
-    compilation by invoking the \c qsb command-line tool at build time. This
-    tool uses QShaderBaker itself and writes the serialized version of the
-    generated QShader into a file. The usage of this class should be
-    restricted to cases where run time compilation cannot be avoided, such as
-    when working with user-provided shader source strings.
+    directly. Rather, all Qt users are encouraged to rely on offline compilation
+    by invoking the \c qsb command-line tool at build time via CMake. The \c qsb
+    tool uses QShaderBaker and writes the serialized version of the generated
+    QShader into a file. The usage of this class should be restricted to cases
+    where run time compilation cannot be avoided, such as when working with
+    user-provided or dynamically generated shader source strings.
 
     The input format is always assumed to be Vulkan-flavored GLSL at the
     moment. See the
@@ -123,7 +134,11 @@ struct QShaderBakerPrivate
     QByteArray preamble;
     int batchLoc = 7;
     bool perTargetEnabled = false;
+    bool breakOnShaderTranslationError = true;
+    QSpirvShader::TessellationInfo tessInfo;
+    QSpirvShader::MultiViewInfo multiViewInfo;
     QShaderBaker::SpirvOptions spirvOptions;
+    QShaderBaker::GlslOptions glslOptions;
     QSpirvCompiler compiler;
     QString errorMessage;
 };
@@ -136,6 +151,16 @@ bool QShaderBakerPrivate::readFile(const QString &fn)
         return false;
     }
     source = f.readAll();
+
+    // Remove UTF-8 BOM, if present. glslang fails otherwise, since the BOM is
+    // removed in glslangValidator (the command-line frontend), not in the
+    // compiler itself. Replicate that here.
+    if (source.size() >= 3) {
+        const unsigned char *p = reinterpret_cast<const unsigned char *>(source.constData());
+        if (p[0] == 0xEF && p[1] == 0xBB && p[2] == 0xBF)
+            source.remove(0, 3);
+    }
+
     sourceFileName = fn;
     return true;
 }
@@ -171,6 +196,11 @@ QShaderBaker::~QShaderBaker()
     \li \c{.geom} - geometry shader
     \li \c{.comp} - compute shader
     \endlist
+
+    \warning \a fileName is expected to contain trusted content. Application
+    developers are advised to carefully consider the potential implications
+    before passing in user-provided source files that are not part of the
+    application.
  */
 void QShaderBaker::setSourceFileName(const QString &fileName)
 {
@@ -200,6 +230,11 @@ void QShaderBaker::setSourceFileName(const QString &fileName)
     Sets the name of the shader source file to \a fileName. This is the file
     that will be read when calling bake(). The shader stage is specified by \a
     stage.
+
+    \warning \a fileName is expected to contain trusted content. Application
+    developers are advised to carefully consider the potential implications
+    before passing in user-provided source files that are not part of the
+    application.
  */
 void QShaderBaker::setSourceFileName(const QString &fileName, QShader::Stage stage)
 {
@@ -211,6 +246,11 @@ void QShaderBaker::setSourceFileName(const QString &fileName, QShader::Stage sta
     Sets the source \a device. This allows using any QIODevice instead of just
     files. \a stage specifies the shader stage, while the optional \a fileName
     contains a filename that is used in the error messages.
+
+    \warning \a device is expected to contain trusted content. Application
+    developers are advised to carefully consider the potential implications
+    before passing in user-provided data from sources that are not under the
+    application's control.
  */
 void QShaderBaker::setSourceDevice(QIODevice *device, QShader::Stage stage, const QString &fileName)
 {
@@ -221,6 +261,11 @@ void QShaderBaker::setSourceDevice(QIODevice *device, QShader::Stage stage, cons
     Sets the input shader \a sourceString. \a stage specified the shader stage,
     while the optional \a fileName contains a filename that is used in the
     error messages.
+
+    \warning \a sourceString is expected to contain trusted content. Application
+    developers are advised to carefully consider the potential implications
+    before passing in user-provided data from sources that are not under the
+    application's control.
  */
 void QShaderBaker::setSourceString(const QByteArray &sourceString, QShader::Stage stage, const QString &fileName)
 {
@@ -380,11 +425,114 @@ void QShaderBaker::setPerTargetCompilation(bool enable)
     d->perTargetEnabled = enable;
 }
 
+/*!
+    Controls the behavior when shader translation (from SPIR-V to
+    GLSL/HLSL/MSL) fails. By default this setting is true, which will cause
+    bake() to return with an error if a requested shader cannot be generated.
+    If that is not desired, and the intention is to generate what we can but
+    silently skip the rest, then set \a enable to false.
+
+    Targeting multiple GLSL versions can lead to errors when a feature is not
+    translatable to a given version. For example, attempting to translate a
+    shader using textureSize() to GLSL ES 100 would fail the entire bake() call
+    with the error message "textureSize is not supported in ESSL 100". If it is
+    acceptable to not have a GLSL ES 100 shader in the result, even though it
+    was requested, then setting this flag to false makes bake() to succeed.
+ */
+void QShaderBaker::setBreakOnShaderTranslationError(bool enable)
+{
+    d->breakOnShaderTranslationError = enable;
+}
+
+/*!
+    When generating MSL shader code for a tessellation control shader, the
+    tessellation \a mode (triangles or quads) must be known upfront. In GLSL
+    this is declared in the tessellation evaluation shader typically, but for
+    Metal it must be known also when generating the compute shader from the
+    tessellation control shader.
+
+    When not set, the default is triangles.
+ */
+void QShaderBaker::setTessellationMode(QShaderDescription::TessellationMode mode)
+{
+    d->tessInfo.infoForTesc.mode = mode;
+}
+
+/*!
+    When generating MSL shader code for a tessellation evaluation shader, the
+    output vertex \a count of the tessellation control shader must be known
+    upfront. in GLSL this would be declared in the tessellation control shader
+    typically, but for Metal it must be known also when generating the vertex
+    shader from the teselation evaluation shader.
+
+    When not set, the default value is 3.
+ */
+void QShaderBaker::setTessellationOutputVertexCount(int count)
+{
+    d->tessInfo.infoForTese.vertexCount = count;
+}
+
+/*!
+    When transpiling shaders using multiview (e.g. a vertex shader using
+    gl_ViewIndex for a renderer relying on GL_OVR_multiview2, VK_KHR_multiview,
+    etc.), for some of the targets it is necessary to declare the number of
+    views in the shader. This is not done in the Vulkan-style GLSL code, and is
+    not relevant for targets such as SPIR-V or HLSL, but is required for OpenGL
+    and GLSL, and so the value has to be provided as additional metadata.
+
+    By default the value is 0, which disables injecting the \c{num_views}
+    statement. Setting 1 is not useful since that is the default \c{num_views}
+    regardless. Therefore \a count should be >= 2 to make an effect. When set
+    to, for example, 2, the generated GLSL shader will contain a
+    \c{layout(num_views = 2) in;} statement.
+
+    Setting a \a count of 2 or greater also injects some preprocessor
+    statements:
+    \c{QSHADER_VIEW_COUNT} is set to \a count, whereas the
+    \c GL_EXT_multiview extension is enabled automatically. Therefore, setting
+    the appropriate
+    \a count can be relevant with other types of shaders as well, e.g. when
+    sharing a uniform buffer between the vertex and fragment shader and both
+    shaders have to be able to write something like
+    \c{#if QSHADER_VIEW_COUNT >= 2}.
+
+    \since 6.7
+ */
+void QShaderBaker::setMultiViewCount(int count)
+{
+    d->multiViewInfo.viewCount = count >= 2 ? count : 0;
+}
+
+/*!
+    \enum QShaderBaker::SpirvOption
+    \value GenerateFullDebugInfo Generate and store additional debug information in the SPIR-V binary.
+    \value StripDebugAndVarInfo Strip all debug and variable name information from the SPIR-V binary.
+ */
+
+/*!
+    Sets additional \a options for the generated SPIR-V binary.
+    By default no flags are set.
+ */
 void QShaderBaker::setSpirvOptions(SpirvOptions options)
 {
     d->spirvOptions = options;
 }
 
+/*!
+    \enum QShaderBaker::GlslOption
+    \value GlslEsFragDefaultFloatPrecisionMedium Emit \c{precision mediump float;} in fragment shaders for GLSL ES.
+ */
+
+/*!
+    Sets additional \a options for the generated GLSL and GLSL ES sources.
+    By default no flags are set.
+    \since 6.9
+ */
+void QShaderBaker::setGlslOptions(GlslOptions options)
+{
+    d->glslOptions = options;
+}
+
 inline size_t qHash(const QShaderBaker::GeneratedShader &k, size_t seed = 0)
 {
     return qHash(k.first, seed) ^ k.second.version();
@@ -487,15 +635,24 @@ QShader QShaderBaker::bake()
         return true;
     };
 
+    QByteArray preamble = d->preamble;
+    if (d->multiViewInfo.viewCount >= 2) {
+        if (d->stage == QShader::VertexStage)
+            preamble += QByteArrayLiteral("\n#extension GL_EXT_multiview : require\n");
+        preamble += QByteArrayLiteral("\n#define QSHADER_VIEW_COUNT ");
+        preamble += QByteArray::number(d->multiViewInfo.viewCount);
+        preamble += QByteArrayLiteral("\n");
+    }
+
     if (!d->perTargetEnabled) {
-        d->compiler.setPreamble(d->preamble);
+        d->compiler.setPreamble(preamble);
         if (!compileSpirvAndBatchable({ QShader::SpirvShader, {} }))
             return QShader();
     } else {
         // per-target compilation. the value here comes from the varying
         // preamble (and so preprocessor defines)
         for (GeneratedShader req: d->reqVersions) {
-            d->compiler.setPreamble(d->preamble + d->perTargetDefines(req));
+            d->compiler.setPreamble(preamble + d->perTargetDefines(req));
             if (!compileSpirvAndBatchable(req))
                 return QShader();
         }
@@ -514,28 +671,44 @@ QShader QShaderBaker::bake()
     // hardcoded rule to pick one), but cannot differ for targets (in
     // per-target mode, hence we can just pick the first SPIR-V binary and
     // generate the reflection data based on that)
-    spirvShader.setSpirvBinary(spirv.constKeyValueBegin()->second);
+    spirvShader.setSpirvBinary(spirv.constKeyValueBegin()->second, d->stage);
     if (batchableSpirv.isEmpty()) {
         bs.setDescription(spirvShader.shaderDescription());
     } else {
-        batchableSpirvShader.setSpirvBinary(batchableSpirv.constKeyValueBegin()->second);
+        batchableSpirvShader.setSpirvBinary(batchableSpirv.constKeyValueBegin()->second, d->stage);
         // prefer the batchable's reflection info with _qt_order and such present
         bs.setDescription(batchableSpirvShader.shaderDescription());
     }
 
     for (const GeneratedShader &req: d->reqVersions) {
         for (const QShader::Variant &v : d->variants) {
-            if (v == QShader::BatchableVertexShader && d->stage != QShader::VertexStage)
-                continue;
+            if (d->stage != QShader::VertexStage) {
+                if (v == QShader::BatchableVertexShader
+                        || v == QShader::UInt32IndexedVertexAsComputeShader
+                        || v == QShader::UInt16IndexedVertexAsComputeShader
+                        || v == QShader::NonIndexedVertexAsComputeShader)
+                {
+                    continue;
+                }
+            }
+            if (req.first != QShader::MslShader && req.first != QShader::MetalLibShader) {
+                if (v == QShader::UInt32IndexedVertexAsComputeShader
+                        || v == QShader::UInt16IndexedVertexAsComputeShader
+                        || v == QShader::NonIndexedVertexAsComputeShader)
+                {
+                    continue;
+                }
+            }
+
             QSpirvShader *currentSpirvShader = nullptr;
             if (d->perTargetEnabled) {
                 // This is expensive too, in addition to the multiple
                 // compilation rounds, but opting in to per-target mode is a
                 // careful, conscious choice (hopefully), so it's fine.
                 if (v == QShader::BatchableVertexShader)
-                    batchableSpirvShader.setSpirvBinary(batchableSpirv[req]);
+                    batchableSpirvShader.setSpirvBinary(batchableSpirv[req], d->stage);
                 else
-                    spirvShader.setSpirvBinary(spirv[req]);
+                    spirvShader.setSpirvBinary(spirv[req], d->stage);
             }
             if (v == QShader::BatchableVertexShader)
                 currentSpirvShader = &batchableSpirvShader;
@@ -550,7 +723,7 @@ QShader QShaderBaker::bake()
             case QShader::SpirvShader:
                 if (d->spirvOptions.testFlag(QShaderBaker::SpirvOption::StripDebugAndVarInfo)) {
                     QString errorMsg;
-                    const QByteArray strippedSpirv = currentSpirvShader->remappedSpirvBinary(QSpirvShader::StripOnly, &errorMsg);
+                    const QByteArray strippedSpirv = currentSpirvShader->remappedSpirvBinary(QSpirvShader::RemapFlag::StripOnly, &errorMsg);
                     if (strippedSpirv.isEmpty()) {
                         d->errorMessage = errorMsg;
                         return QShader();
@@ -563,13 +736,25 @@ QShader QShaderBaker::bake()
             case QShader::GlslShader:
             {
                 QSpirvShader::GlslFlags flags;
-                if (req.second.flags().testFlag(QShaderVersion::GlslEs))
-                    flags |= QSpirvShader::GlslEs;
+                if (req.second.flags().testFlag(QShaderVersion::GlslEs)) {
+                    flags |= QSpirvShader::GlslFlag::GlslEs;
+                    if (d->glslOptions.testFlag(GlslOption::GlslEsFragDefaultFloatPrecisionMedium))
+                        flags |= QSpirvShader::GlslFlag::EsFragDefaultFloatPrecisionMedium;
+                }
                 QVector<QSpirvShader::SeparateToCombinedImageSamplerMapping> separateToCombinedImageSamplerMappings;
-                shader.setShader(currentSpirvShader->translateToGLSL(req.second.version(), flags, &separateToCombinedImageSamplerMappings));
+                shader.setShader(currentSpirvShader->translateToGLSL(req.second.version(),
+                                                                     flags,
+                                                                     d->stage,
+                                                                     d->multiViewInfo,
+                                                                     &separateToCombinedImageSamplerMappings));
                 if (shader.shader().isEmpty()) {
-                    d->errorMessage = currentSpirvShader->translationErrorMessage();
-                    return QShader();
+                    if (d->breakOnShaderTranslationError) {
+                        d->errorMessage = currentSpirvShader->translationErrorMessage();
+                        return QShader();
+                    } else {
+                        d->errorMessage += QLatin1String(" ") + currentSpirvShader->translationErrorMessage();
+                        continue;
+                    }
                 }
                 if (!separateToCombinedImageSamplerMappings.isEmpty()) {
                     const QShaderDescription desc = bs.description();
@@ -602,8 +787,13 @@ QShader QShaderBaker::bake()
                 QShader::NativeResourceBindingMap nativeBindings;
                 shader.setShader(currentSpirvShader->translateToHLSL(req.second.version(), &nativeBindings));
                 if (shader.shader().isEmpty()) {
-                    d->errorMessage = currentSpirvShader->translationErrorMessage();
-                    return QShader();
+                    if (d->breakOnShaderTranslationError) {
+                        d->errorMessage = currentSpirvShader->translationErrorMessage();
+                        return QShader();
+                    } else {
+                        d->errorMessage += QLatin1String(" ") + currentSpirvShader->translationErrorMessage();
+                        continue;
+                    }
                 }
                 bs.setResourceBindingMap(key, nativeBindings);
             }
@@ -611,13 +801,44 @@ QShader QShaderBaker::bake()
             case QShader::MslShader:
             {
                 QShader::NativeResourceBindingMap nativeBindings;
-                shader.setShader(currentSpirvShader->translateToMSL(req.second.version(), &nativeBindings));
+                QShader::NativeShaderInfo shaderInfo;
+                QSpirvShader::MslFlags flags;
+                if (d->stage == QShader::VertexStage) {
+                    switch (v) {
+                    case QShader::UInt16IndexedVertexAsComputeShader:
+                        flags |= QSpirvShader::MslFlag::VertexAsCompute;
+                        flags |= QSpirvShader::MslFlag::WithUInt16Index;
+                        break;
+                    case QShader::UInt32IndexedVertexAsComputeShader:
+                        flags |= QSpirvShader::MslFlag::VertexAsCompute;
+                        flags |= QSpirvShader::MslFlag::WithUInt32Index;
+                        break;
+                    case QShader::NonIndexedVertexAsComputeShader:
+                        flags |= QSpirvShader::MslFlag::VertexAsCompute;
+                        break;
+                    default:
+                        break;
+                    }
+                }
+                shader.setShader(currentSpirvShader->translateToMSL(req.second.version(),
+                                                                    flags,
+                                                                    d->stage,
+                                                                    &nativeBindings,
+                                                                    &shaderInfo,
+                                                                    d->multiViewInfo,
+                                                                    d->tessInfo));
                 if (shader.shader().isEmpty()) {
-                    d->errorMessage = currentSpirvShader->translationErrorMessage();
-                    return QShader();
+                    if (d->breakOnShaderTranslationError) {
+                        d->errorMessage = currentSpirvShader->translationErrorMessage();
+                        return QShader();
+                    } else {
+                        d->errorMessage += QLatin1String(" ") + currentSpirvShader->translationErrorMessage();
+                        continue;
+                    }
                 }
                 shader.setEntryPoint(QByteArrayLiteral("main0"));
                 bs.setResourceBindingMap(key, nativeBindings);
+                bs.setNativeShaderInfo(key, shaderInfo);
             }
                 break;
             default:
